package ca.recrown.islandsurvivalcraft.world.generation;

import java.util.LinkedList;
import java.util.Random;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import org.bukkit.Chunk;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.block.Biome;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.world.ChunkLoadEvent;
import org.bukkit.generator.ChunkGenerator;

import ca.recrown.islandsurvivalcraft.Utilities;
import ca.recrown.islandsurvivalcraft.caching.Cache;
import ca.recrown.islandsurvivalcraft.datatypes.Point2;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;
import ca.recrown.islandsurvivalcraft.world.shaders.WorldHeightShader;

public class IslandWorldChunkGenerator extends ChunkGenerator implements Listener {
    private final Cache<Point2, Double> blockValueCache = new Cache<>(262144);
    private final Cache<Point2, Biome[]> biomeCache = new Cache<>(262144);
    private final Cache<Point2, Boolean> chunkExistenceCache = new Cache<>(32768);
    private final BiomeSelector biomeSelector = new BiomeSelector();
    private final ExecutorService exAlpha = Utilities.ISC_EXECUTOR_ALPHA;
    private final ExecutorService exBeta = Utilities.ISC_EXECUTOR_BETA;
    private volatile World currentWorld;

    public IslandWorldChunkGenerator() {
        biomeSelector.initialize();
    }

    @Override
    public ChunkData generateChunkData(World world, Random random, int chunkX, int chunkZ, BiomeGrid biomeGrid) {
        this.currentWorld = world;
        IslandWorldMapper mapper = new IslandWorldMapper(random, blockValueCache);
        TemperatureMapGenerator temperatureMapGenerator = new TemperatureMapGenerator(world.getSeed());
        WorldHeightShader heightShader = new WorldHeightShader(random, mapper, world.getSeaLevel(), world.getMaxHeight(), 3);
        BiomeGenerator biomeGenerator = new UniBiomeIslandGenerator();

        int maxHeight = world.getMaxHeight();
        LinkedList<Future<Boolean>> tasks = new LinkedList<>();
        ChunkData chunkData = createChunkData(world);
        Future<Boolean> preLoader = exAlpha.submit(() -> {
            for (int x = Utilities.CHUNK_SIZE - 1; x >= 0; x--) {
                for (int z = Utilities.CHUNK_SIZE - 1; z >= 0; z--) {
                    if (Thread.currentThread().isInterrupted()) return false;
                    mapper.getWorldBlockValue(Utilities.CHUNK_SIZE * chunkX + x, Utilities.CHUNK_SIZE * chunkZ + z);
                }
            }
            return true;
        });

        Biome[][] biomes = new Biome[Utilities.CHUNK_SIZE][Utilities.CHUNK_SIZE];
        for (int x = 0; x < Utilities.CHUNK_SIZE; x++) {
            for (int z = 0; z < Utilities.CHUNK_SIZE; z++) {
                final int localX = x;
                final int localZ = z;
                if (biomes[localX][localZ] == null) {
                    biomeGenerator.generateBiomeColumn(biomes, world, chunkX, chunkZ, localX, localZ, mapper, biomeSelector,
                            temperatureMapGenerator, biomeCache, chunkExistenceCache);
                }
                if (biomes[localX][localZ] == null) throw new IllegalStateException("Biome was null.");
                tasks.add(exBeta.submit(() -> {
                    for (int y = 0; y < maxHeight; y++) {
                        biomeGrid.setBiome(localX, y, localZ, biomes[localX][localZ]);
                    }
                    return true;
                }));

                final int worldX = Utilities.CHUNK_SIZE * chunkX + localX;
                final int worldZ = Utilities.CHUNK_SIZE * chunkZ + localZ;
                int height = heightShader.getAltitude(worldX, worldZ, biomes[localX][localZ]);
                chunkData.setRegion(localX, 1, localZ, localX + 1, height, localZ + 1, Material.DIAMOND_BLOCK);
            }
        }
        chunkData.setRegion(0, 0, 0, 16, 1, 16, Material.BEDROCK);

        preLoader.cancel(false);
        try {
            while (!tasks.isEmpty()) {
                tasks.poll().get();
            }
        } catch (ExecutionException | InterruptedException e) {
            e.printStackTrace();
            throw new IllegalStateException(e.getCause().getMessage());
        }
        return chunkData;
    }
    
    @Override
    public boolean canSpawn(World world, int x, int z) {
        return super.canSpawn(world, x, z);
    }

    @Override
    public boolean isParallelCapable() {
        return true;
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onChunkLoaded(ChunkLoadEvent event) {
        if (event.isNewChunk()) {
            if (currentWorld == event.getWorld()) {
                Chunk chunk = event.getChunk();
                
                chunkExistenceCache.set(new Point2(chunk.getX(), chunk.getZ()), true);
            }
        }
    }
}